import 'dart:async';
import 'dart:collection';

import 'package:aitrainer_app/bloc/menu/menu_bloc.dart';
import 'package:aitrainer_app/model/cache.dart';
import 'package:aitrainer_app/model/exercise_ability.dart';
import 'package:aitrainer_app/model/exercise_plan.dart';
import 'package:aitrainer_app/model/exercise_plan_detail.dart';
import 'package:aitrainer_app/model/exercise_plan_template.dart';
import 'package:aitrainer_app/model/exercise_type.dart';
import 'package:aitrainer_app/repository/workout_tree_repository.dart';
import 'package:aitrainer_app/service/exercise_plan_service.dart';
import 'package:aitrainer_app/util/enums.dart';
import 'package:aitrainer_app/util/track.dart';
import 'package:bloc/bloc.dart';
import 'package:equatable/equatable.dart';

part 'test_set_edit_event.dart';
part 'test_set_edit_state.dart';

class TestSetEditBloc extends Bloc<TestSetEditEvent, TestSetEditState> {
  final String templateName;
  final String templateNameTranslation;
  late String templateDescription;
  final WorkoutTreeRepository workoutTreeRepository;
  final MenuBloc menuBloc;
  final List<ExerciseType> _exerciseTypes = [];
  final List<ExerciseType> _actualExerciseTypes = [];
  final HashMap<int, ExerciseType?> _exercisePlanDetails = HashMap();

  TestSetEditBloc(
      {required this.templateName, required this.templateNameTranslation, required this.workoutTreeRepository, required this.menuBloc})
      : super(TestSetEditInitial()) {
    if (Cache().exercisePlanTemplates.isNotEmpty) {
      Cache().exercisePlanTemplates.forEach((element) {
        final ExercisePlanTemplate template = element as ExercisePlanTemplate;
        if (template.name == templateName) {
          templateDescription = template.descriptionTranslation;
          template.exerciseTypes.forEach((id) {
            final ExerciseType? exerciseType = Cache().getExerciseTypeById(id);
            if (exerciseType != null) {
              _exerciseTypes.add(exerciseType);
              _actualExerciseTypes.add(exerciseType);
              _exercisePlanDetails[exerciseType.exerciseTypeId] = exerciseType;
            }
          });
        }
      });
    }
  }

  @override
  Stream<TestSetEditState> mapEventToState(TestSetEditEvent event) async* {
    try {
      if (event is TestSetEditChangeExerciseType) {
        yield TestSetEditLoading();
        final List<ExerciseType> alternatives = workoutTreeRepository.getExerciseTypeAlternatives(event.exerciseTypeId);
        final ExerciseType? exerciseType = Cache().getExerciseTypeById(event.exerciseTypeId);
        if (exerciseType != null) {
          if (_exercisePlanDetails[event.exerciseTypeId] == null) {
            /// it was skipped
            _exercisePlanDetails[exerciseType.exerciseTypeId] = exerciseType;
          } else {
            if (event.index == 0) {
              _exercisePlanDetails[exerciseType.exerciseTypeId] = exerciseType;
            } else {
              final changedExerciseType = alternatives[event.index - 1];
              _exercisePlanDetails[exerciseType.exerciseTypeId] = changedExerciseType;
            }
          }
        }

        // to keep the slider accurate
        refreshActualPlan();
        yield TestSetEditReady();
      } else if (event is TestSetEditDeleteExerciseType) {
        yield TestSetEditLoading();
        final ExerciseType? exerciseType = Cache().getExerciseTypeById(event.exerciseTypeId);
        if (exerciseType != null) {
          _exercisePlanDetails[exerciseType.exerciseTypeId] = null;
        }
        refreshActualPlan();
        yield TestSetEditReady();
      } else if (event is TestSetEditSubmit) {
        yield TestSetEditLoading();
        ExercisePlan exercisePlan = ExercisePlan(templateNameTranslation, Cache().userLoggedIn!.customerId!);
        exercisePlan.private = true;
        exercisePlan.type = ExerciseAbility.mini_test_set.enumToString();
        exercisePlan.dateAdd = DateTime.now();
        ExercisePlan savedExercisePlan = await ExercisePlanApi().saveExercisePlan(exercisePlan);

        List<ExercisePlanDetail> details = [];
        for (var entry in _exercisePlanDetails.entries) {
          if (entry.value != null) {
            ExercisePlanDetail exercisePlanDetail = ExercisePlanDetail(entry.value!.exerciseTypeId);
            exercisePlanDetail.exercisePlanId = savedExercisePlan.exercisePlanId!;
            exercisePlanDetail.serie = 1;

            ExercisePlanDetail savedDetail = await ExercisePlanApi().saveExercisePlanDetail(exercisePlanDetail);
            exercisePlanDetail.exercisePlanDetailId = savedDetail.exercisePlanDetailId;
            exercisePlanDetail.exercises = [];
            details.add(exercisePlanDetail);
          }
        }
        Cache().saveActiveExercisePlan(exercisePlan, details);
        Track().track(TrackingEvent.test_set_edit, eventValue: templateName);
        yield TestSetEditSaved();
      }
    } on Exception catch (e) {
      yield TestSetEditError(message: e.toString());
    }
  }

  List get exerciseTypes => this._exerciseTypes;
  List get actualExerciseTypes => this._actualExerciseTypes;
  HashMap get exercisePlanDetails => this._exercisePlanDetails;

  void refreshActualPlan() {
    _actualExerciseTypes.removeRange(0, _actualExerciseTypes.length - 1);
    _exercisePlanDetails.keys.forEach((element) {
      final ExerciseType? exerciseType = Cache().getExerciseTypeById(element);
      if (exerciseType != null) {
        _actualExerciseTypes.add(exerciseType);
      }
    });
  }
}
